#include <os.h>
#include "navigation.h"
#include "output.h"

inline int prevnewl(char* textbuffer, int pos){//returns position_of_last_newline_character_before_pos + 1 or zero if not found
    pos--;
    while(textbuffer[pos] != '\n' && pos >= 0)
        pos--;
    return ++pos;
}

int getw(char* textbuffer, int pos){
    int tmppos = prevnewl(textbuffer, pos);
    int w = 0;
    while(tmppos <= pos){
        for(w = 0; textbuffer[tmppos] != '\n' && w < SCREEN_WIDTH/CHAR_WIDTH; tmppos++){
            if(tmppos >= pos)//how many characters into the line is pos?
                break;
            if(textbuffer[tmppos] == '\t')
                w += TAB_WIDTH - (w % TAB_WIDTH);
            else
                w++;
        }
        if(textbuffer[tmppos] == '\n'){
            if(tmppos >= pos)
                break;
            tmppos++;
        }
        if(tmppos >= pos)
            break;
    }
    if(w >= SCREEN_WIDTH/CHAR_WIDTH)
        return 0;
    return w;
}

int getnum(char* textbuffer, int pos){
    int tmppos = prevnewl(textbuffer, pos);
    int w = 0;
    int num = 0;
    while(tmppos <= pos){
        num = 0;
        for(w = 0; textbuffer[tmppos] != '\n' && w < SCREEN_WIDTH/CHAR_WIDTH; tmppos++){
            if(tmppos >= pos)//how many characters into the line is pos?
                break;
            if(textbuffer[tmppos] == '\t')
                w += TAB_WIDTH - (w % TAB_WIDTH);
            else
                w++;
            num++;
        }
        if(textbuffer[tmppos] == '\n'){
            if(tmppos >= pos)
                break;
            tmppos++;
        }
        if(tmppos >= pos)
            break;
    }
    if(w >= SCREEN_WIDTH/CHAR_WIDTH)
        return 0;
    return num;
}

inline int currentline(char* textbuffer, int pos){
    return pos - getnum(textbuffer, pos);
}

int prevline(char* textbuffer, int pos){//if in first line, returns 0
    int tmppos;
    int w;
    int linestart = 0;
    pos = currentline(textbuffer, pos);
    if(pos == 0)
        return 0;
    pos--;
    tmppos = prevnewl(textbuffer, pos);
    while(tmppos <= pos){
        linestart = tmppos;//save start of line
        for(w = 0; textbuffer[tmppos] != '\n' && w < SCREEN_WIDTH/CHAR_WIDTH && textbuffer[tmppos] != '\0'; tmppos++){
            if(textbuffer[tmppos] == '\t')
                w += TAB_WIDTH - (w % TAB_WIDTH);
            else
                w++;
        }
        if((textbuffer[tmppos] == '\n' || textbuffer[tmppos] == '\0') && w < SCREEN_WIDTH/CHAR_WIDTH)
            tmppos++;
    }
    return linestart;
}

int checkednextline(char* textbuffer, int pos){//if in last line, returns 0
    int tmppos = prevnewl(textbuffer, pos);//start at first character after newline
    int w;
    while(tmppos <= pos){
        for(w = 0; textbuffer[tmppos] != '\n' && w < SCREEN_WIDTH/CHAR_WIDTH && textbuffer[tmppos] != '\0'; tmppos++){
            if(textbuffer[tmppos] == '\t')
                w += TAB_WIDTH - (w % TAB_WIDTH);
            else
                w++;
        }
        if(textbuffer[tmppos] == '\0')//return 0 if in last line
            return 0;
        if(textbuffer[tmppos] == '\n' && w < SCREEN_WIDTH/CHAR_WIDTH)
            tmppos++;
    }
    return tmppos;
}

int nextline(char* textbuffer, int pos){//if in last line, returns last line again
    if(checkednextline(textbuffer, pos) == 0)
        return currentline(textbuffer, pos);
    return checkednextline(textbuffer, pos);
}

int gotow(char* textbuffer, int pos, int w){//returns position of char at width w of the current line
    pos = currentline(textbuffer, pos);
    int tmpw;
    for(tmpw = 0; tmpw < w; pos++){
        if(textbuffer[pos] == '\n' || textbuffer[pos] == '\0')
            break;
        else if(textbuffer[pos] == '\t')
            tmpw += TAB_WIDTH - (tmpw % TAB_WIDTH);
        else
            tmpw++;
    }
    return pos;
}

int go_down(char* textbuffer, int pos){
    int w = getw(textbuffer, pos);
    pos = nextline(textbuffer, pos);
    return gotow(textbuffer, pos, w);
}

int go_up(char* textbuffer, int pos){
    int w = getw(textbuffer, pos);
    pos = prevline(textbuffer, pos);
    return gotow(textbuffer, pos, w);
}
