#include <os.h>
#include "input.h"
#include "output.h"
#include "utils.h"
#include "menu.h"
#include "fileio.h"
#include "navigation.h"

#define CHARDELAY 60//=600ms
#define CURSORDELAY 20

#define DELAYFIRST 60
#define DELAYAFTER 6

/*
    - selection bei ctrl navigs testen
    - scrolling bei ctrl navigs konsistent mit os machen
    - cursorscreenfreie lsung? damit nicht probleme wie bei multilines
    - berichteter bug (mdr1) crash beim ffnen von datei
    - cursorscreenrow-rechnung bei paste, cut, delete einer auswahl neu schreiben
    - schnelles scrolling
    - mehrere text in tabs
    - gui improvements
    - beschleunigender input: beim ersten mal ~500ms delay, dann nur noch ~50ms -> cursordelay&chardelay vereinheitlichen
    - warnung falls an anderem ort im filesystem als in /documents/ -> falls .. im path
    - getline sichern
    - malloc sichern
    - main bersichtlicher machen
    - vl optim: falls keine nderungen nichts dispen
    - bei esc nur meldung anzeigen falls nderung seit letztem speichern
    - nspireio anschauen: vl besser geeignet
    - disp optimieren: fps sinkt von 100 auf 35 bei vollem bildschirm (timer statt sleep wrde symptome auch beheben)
    - windows-style newline (\r \n) mindestens erkennen und optional benutzen
    - scrollbalken
    - scrolling vl. mit swipen, oder so
    - option zum automatischen einfgen von leerzeichen statt tabs
    - warnung falls datei schon existiert -> berschrieben wird
    - sonderzeichenpopup
    - file browser
    - nsdl/nrgblib anschauen vl.
    - ist editieren von files die NULL enthalten ntig? wrde ntig machen mem... statt str... und va ____andere kontrolle der bufferlnge_____
    - undo,redo
    - automatisches anhngen von .txt.tns an path verbessern: zb keine .txt.txt.tns produzieren
    - wre verkleinerung des allozierten speichers ntig wenn grosse bereiche gelscht?
*/

/*
mem:    0   |   1   |   2   |   ...
pos: 0      1       2       3       ...
*/

int main(int argc, char* argv[]){
    uint8_t* scrbuf = (uint8_t*) malloc(SCREEN_BYTES_SIZE);
    if(scrbuf == NULL)
        return 1;
    memset(scrbuf,0xFF,SCREEN_BYTES_SIZE);

    char* tmpcp;//char pointer used in various places

    FILE* file = NULL;
    char* path;
    path = argv[0];
    tmpcp = strrchr(path,'/') + 1;
    *tmpcp = '\0';
    char savepath[256];
    savepath[0] = '\0';

    char* textbuffer = malloc(1024);
    if(textbuffer == NULL){
        free(scrbuf);
        return 1;
    }
    *textbuffer = '\0';

    char inpbuf = '\0';
    char blockkey = '\0';
    int chardelaycounter = CHARDELAY;

    uint8_t arrowbuf = 0;
    uint8_t blockarrow = 0;
    int cursordelaycounter = CURSORDELAY;

    int pos = 0;
    int selectionstart = 0;
    int selectionend = 0;

    int cursorscreenrow = 0;
    char* displinep;
    BOOL menu = FALSE;

    int selecteditem = 1;
    int lastitem;

    char* clipboard = NULL;

    if(argc >= 2){//if opened through file association
        strcpy(savepath, argv[1]);
        file = fopen(savepath,"rb");
        if(file == NULL){
            free(textbuffer);
            free(scrbuf);
            return 1;
        }
        fseek(file, 0, SEEK_END);
        int filelen = ftell(file);
        rewind(file);
        free(textbuffer);
        textbuffer = malloc(filelen + 1024 - (filelen%1024));
        if(textbuffer == NULL){
            free(scrbuf);
            return 1;
        }
        fread(textbuffer, 1, filelen, file);
        fclose(file);
        printf("%i",filelen);
        textbuffer[filelen] = '\0';
    }

    int fps = 0;
    uint32_t oldtime = 0;

    while(1){
        if(oldtime != *(uint32_t*)0x90090000){
            oldtime = *(uint32_t*)0x90090000;
            printf("fps:%i\n",fps);
            fps = 0;
        }
        ++fps;

        sleep(10);//->100fps

        //disp in both menu and text input mode:
        memcpy(SCREEN_BASE_ADDRESS, scrbuf, SCREEN_BYTES_SIZE);
        memset(scrbuf, 0xFF, SCREEN_BYTES_SIZE);
        displinep = getline(textbuffer, pos, cursorscreenrow);
        dispStringWithSelection(scrbuf, 0, 0, displinep, selectionstart - (displinep - textbuffer), selectionend - (displinep - textbuffer));
        //dispString(scrbuf, 0, 0, displinep);
        //dispSelection(scrbuf, selectionstart, selectionend, displinep);

        if(!menu){
            if(on_key_pressed() || isKeyPressed(KEY_NSPIRE_ESC)){
                wait_no_key_pressed();
                unsigned inp = show_msgbox_3b("", "Save Document?", "Yes", "No", "Cancel");
                if(inp == 1){
                    if(strlen(savepath) == 0)
                        file = fopen(strcat(path,"noname.txt.tns"),"ab");
                    else
                        file = fopen(savepath, "wb");
                    fwrite(textbuffer,1,strlen(textbuffer),file);
                    fclose(file);
                    break;
                }
                if(inp == 2)
                    break;
                wait_no_key_pressed();
            }

            if(strlen(textbuffer) % 1024 == 1023){//if buffer full, make it larger
                textbuffer = realloc(textbuffer, strlen(textbuffer) + 1 + 1024);
                if(textbuffer == NULL)
                    break;
            }

            //open menu:
            if(isKeyPressed(KEY_NSPIRE_MENU) || isKeyPressed(KEY_NSPIRE_DOC)){
                menu = TRUE;
                selecteditem = 1;
            }

            //open:
            else if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_O)){
                selectionstart = 0;
                selectionend = 0;
                if(open(file, path, savepath, textbuffer, &pos, &cursorscreenrow) == 1)
                    break;
            }

            //saving:
            else if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_S)){
                save(file, path, savepath, textbuffer);
            }

            //copy:
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_C)){
                if(selectionstart != selectionend){
                    if(clipboard != NULL)
                        free(clipboard);
                    clipboard = (char*) malloc(selectionend - selectionstart + 1);
                    printf("%i\n", selectionend - selectionstart + 1);
                    memcpy(clipboard, textbuffer + selectionstart, selectionend - selectionstart);
                    clipboard[selectionend - selectionstart] = '\0';
                    printf("%s\n", clipboard);
                }
                while(isKeyPressed(KEY_NSPIRE_C))
                    idle();
            }

            //cut:
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_X)){
                if(selectionstart != selectionend){
                    if(clipboard != NULL)
                        free(clipboard);
                    clipboard = (char*) malloc(selectionend - selectionstart + 1);
                    printf("%i\n", selectionend - selectionstart + 1);
                    memcpy(clipboard, textbuffer + selectionstart, selectionend - selectionstart);
                    clipboard[selectionend - selectionstart] = '\0';
                    delselection(textbuffer, selectionstart, selectionend);
                    for(;pos > selectionstart; pos--){
                        if(textbuffer[pos-1] == '\n')
                            cursorscreenrow--;
                    }
                    selectionend = selectionstart;
                    printf("%s\n", clipboard);
                }
                while(isKeyPressed(KEY_NSPIRE_X))
                    idle();
            }

            //paste:
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_V)){
                if(clipboard != NULL){
                    if(selectionstart != selectionend){
                        delselection(textbuffer, selectionstart, selectionend);
                        selectionend = selectionstart;
                        pos = selectionstart;
                    }
                    if(strlen(textbuffer) % 1024 + strlen(clipboard) > 1023)//allocate enough memory
                        textbuffer = realloc(textbuffer, strlen(textbuffer) + strlen(clipboard) + 1024 - (strlen(textbuffer) + strlen(clipboard)) % 1024);
                    insertstr(textbuffer, pos, clipboard);
                    selectionstart += strlen(clipboard);
                    selectionend = selectionstart;
                    for(; pos < selectionstart; pos++){
                        if(textbuffer[pos] == '\n')
                            cursorscreenrow++;
                    }
                }
                while(isKeyPressed(KEY_NSPIRE_V))
                    idle();
            }

            //select all:
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_A)){
                selectionstart = 0;
                selectionend = strlen(textbuffer);
                pos = 0;
                cursorscreenrow = 0;
                while(isKeyPressed(KEY_NSPIRE_A))
                    idle();
            }

            //go to start of file
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_7)){
                selectionstart = 0;
                pos = 0;
                cursorscreenrow = 0;
                if(!isKeyPressed(KEY_NSPIRE_SHIFT))
                    selectionend = 0;
                while(isKeyPressed(KEY_NSPIRE_7))
                    idle();
            }

            //go to end of file
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_1)){
                pos = 0;
                for(cursorscreenrow = 0; checkednextline(textbuffer, pos) != 0 && cursorscreenrow < SCREEN_HEIGHT/CHAR_HEIGHT - 1; cursorscreenrow++)//find out maximal cursorscreenrow
                    pos = checkednextline(textbuffer, pos);
                pos = strchr(textbuffer, '\0') - textbuffer;
                selectionend = pos;
                if(!isKeyPressed(KEY_NSPIRE_SHIFT))
                    selectionstart = pos;
                while(isKeyPressed(KEY_NSPIRE_1))
                    idle();
            }

            //pg up
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_9)){
                int i;
                int oldpos = pos;
                int w = getw(textbuffer, pos);
                for(i = 0; i < SCREEN_HEIGHT / CHAR_HEIGHT; i++)
                    pos = prevline(textbuffer, pos);
                pos = gotow(textbuffer, pos, w);

                if(isKeyPressed(KEY_NSPIRE_SHIFT)){
                    if(pos >= selectionstart)
                        selectionend = pos;
                    else if(oldpos == selectionstart)
                        selectionstart = pos;
                    else{
                        selectionend = selectionstart;
                        selectionstart = pos;
                    }
                }
                else{
                    selectionstart = pos;
                    selectionend = pos;
                }
                while(isKeyPressed(KEY_NSPIRE_9))
                    idle();
            }

            //pg down
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_3)){
                int i;
                int oldpos = pos;
                int w = getw(textbuffer, pos);
                for(i = 0; i < SCREEN_HEIGHT / CHAR_HEIGHT; i++)
                    pos = nextline(textbuffer, pos);
                pos = gotow(textbuffer, pos, w);

                if(isKeyPressed(KEY_NSPIRE_SHIFT)){
                    if(pos <= selectionend)
                        selectionstart = pos;
                    else if(oldpos == selectionend)
                        selectionend = pos;
                    else{
                        selectionstart = selectionend;
                        selectionend = pos;
                    }
                }
                else{
                    selectionstart = pos;
                    selectionend = pos;
                }
                while(isKeyPressed(KEY_NSPIRE_3))
                    idle();
            }

            /*
            //home
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_...)){
                pos = currentline(textbuffer, pos);
                selectionstart = pos;
                if(!isKeyPressed(KEY_NSPIRE_SHIFT))
                    selectionend = pos;
                while(isKeyPressed(KEY_NSPIRE_...))
                    idle();
            }

            //end
            if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_...)){
                pos = nextline(textbuffer, pos) - 1;//find alternative in last line
                selectionend = pos;
                if(!isKeyPressed(KEY_NSPIRE_SHIFT))
                    selectionstart = pos;
                while(isKeyPressed(KEY_NSPIRE_...))
                    idle();
            }
            */

            //input(alphanum/newline/tab/del):
            if(!isKeyPressed(KEY_NSPIRE_CTRL) && (inpbuf = readc()) != '\0'){
                if(inpbuf == blockkey && chardelaycounter)//if key hasn't been released and repeated keypress delay isn't over, ignore keypress
                    inpbuf = '\0';
                else{
                    blockkey = inpbuf;
                    chardelaycounter = CHARDELAY;
                    if(selectionstart != selectionend){
                        delselection(textbuffer, selectionstart, selectionend);
                        /*for(;pos > selectionstart; pos--){
                            if(textbuffer[pos-1] == '\n')
                                cursorscreenrow--;
                        }*/
                        pos = selectionstart;
                        selectionend = selectionstart;
                    }
                    else if(inpbuf == '\b'){//if selection: already deleted -> do nothing more
                        if(strlen(textbuffer) != 0 && pos != 0){
                            if(delchar(textbuffer, pos) == '\n')
                                cursorscreenrow--;
                            pos--;
                            selectionstart = pos;
                            selectionend = pos;
                        }
                    }
                    if(inpbuf != '\b'){
                        if(inpbuf == '\n')
                            cursorscreenrow++;
                        insert(textbuffer, pos, inpbuf);
                        pos++;
                        selectionstart = pos;
                        selectionend = pos;
                    }
                }
            }
            else//if no key pressed <==> if key released
                blockkey = '\0';
            if(chardelaycounter)
                chardelaycounter--;

            //cursor moving and selecting:
            if((arrowbuf = readarrow()) != 0){
                if(arrowbuf == blockarrow && cursordelaycounter)
                    arrowbuf = 0;
                else{
                    blockarrow = arrowbuf;
                    cursordelaycounter = CURSORDELAY;
                    if(arrowbuf == ARROW_RIGHT_NUM){
                        if(pos < strlen(textbuffer)){
                            if(textbuffer[pos] == '\n')
                                cursorscreenrow++;
                            pos++;
                            if(isKeyPressed(KEY_NSPIRE_SHIFT)){//select text
                                if(pos == selectionend + 1)
                                    selectionend++;
                                else
                                    selectionstart++;
                            }
                            else{
                                selectionstart = pos;
                                selectionend = pos;
                            }
                        }
                    }
                    else if(arrowbuf == ARROW_LEFT_NUM){
                        if(pos > 0){
                            pos--;
                            if(textbuffer[pos] == '\n')
                                cursorscreenrow--;
                            if(isKeyPressed(KEY_NSPIRE_SHIFT)){//select text
                                if(pos == selectionstart - 1)
                                    selectionstart--;
                                else
                                    selectionend--;
                            }
                            else{
                                selectionstart = pos;
                                selectionend = pos;
                            }
                        }
                    }
                    else if(arrowbuf == ARROW_UP_NUM){
                        if(currentline(textbuffer, pos) != 0){//if cursor not in first line of file
                            cursorscreenrow--;
                            pos = go_up(textbuffer, pos);
                            if(isKeyPressed(KEY_NSPIRE_SHIFT)){
                                if(go_down(textbuffer, pos) == selectionstart)
                                    selectionstart = pos;
                                else if(pos > selectionstart)
                                    selectionend = pos;
                                else{
                                    selectionend = selectionstart;
                                    selectionstart = pos;
                                }

                            }
                            else{
                                selectionstart = pos;
                                selectionend = pos;
                            }
                        }
                    }
                    else if(arrowbuf == ARROW_DOWN_NUM){
                        if(checkednextline(textbuffer, pos) != 0){//if cursor not in last line of file
                            cursorscreenrow++;
                            pos = go_down(textbuffer, pos);
                            if(isKeyPressed(KEY_NSPIRE_SHIFT)){
                                if(go_up(textbuffer, pos) == selectionend)
                                    selectionend = pos;
                                else if(pos < selectionend)
                                    selectionstart = pos;
                                else{
                                    selectionstart = selectionend;
                                    selectionend = pos;
                                }
                            }
                            else{
                                selectionstart = pos;
                                selectionend = pos;
                            }
                        }
                    }
                }
            }
            else
                blockarrow = 0;
            if(cursordelaycounter)
                cursordelaycounter--;

            if(cursorscreenrow < 0)
                cursorscreenrow = 0;
            if(cursorscreenrow >= SCREEN_HEIGHT/CHAR_HEIGHT)
                cursorscreenrow = SCREEN_HEIGHT/CHAR_HEIGHT - 1;

            //disp:
            dispCursor(scrbuf, pos-(displinep-textbuffer), displinep);
        }

        //if menu is open:
        else{
            lastitem = dispMenu(scrbuf, 0, 0, "Open    Ctrl+O\nSave    Ctrl+S\nSave as", selecteditem);

            //closing menu:
            if(isKeyPressed(KEY_NSPIRE_ESC) || on_key_pressed()){
                menu = FALSE;
                wait_no_key_pressed();
            }

            //choosing an option:
            if(isKeyPressed(KEY_NSPIRE_ENTER)){
                if(selecteditem == 1){
                    if(open(file, path, savepath, textbuffer, &pos, &cursorscreenrow) == 1)
                        break;
                }
                else if(selecteditem == 2)
                    save(file, path, savepath, textbuffer);
                else if(selecteditem == 3)
                    saveAs(file, path, savepath, textbuffer);

                menu = FALSE;
            }

            //navigating:
            if((arrowbuf = readarrow()) != 0){
                if(arrowbuf == blockarrow && cursordelaycounter)
                    arrowbuf = 0;
                else{
                    blockarrow = arrowbuf;
                    cursordelaycounter = CURSORDELAY;
                    if(arrowbuf == ARROW_UP_NUM){
                        selecteditem--;
                        if(selecteditem == 0)
                            selecteditem = lastitem;
                    }
                    else if(arrowbuf == ARROW_DOWN_NUM){
                        selecteditem++;
                        if(selecteditem > lastitem)
                            selecteditem = 1;
                    }
                }
            }
            else
                blockarrow = 0;
            if(cursordelaycounter)
                cursordelaycounter--;
        }
    }
    free(clipboard);
    free(textbuffer);
    free(scrbuf);
    refresh_osscr();
    return 0;
}
